/* ------------------------------------------------------------------------ 
 *  File: ExtendableClassLoader.java
 * 
 *  DATE         AUTHOR        CHANGES
 *  1996         C. McManis    Created
 *  Jan 28 1998  O. Dedieu     Modify to support class paths adding
 *  Feb 09 1998  M. Travers    Appropriated and lightly modified for Skij.
 *    To do: make it work with ZIPs and JARs
 * ------------------------------------------------------------------------ 
 *
 * This class has been developed from the SimpleClassLoader class
 * proposed by Chuck McManis in JavaWorld' article :
 * http://www.javaworld.com/javaworld/jw-10-indepth.html
 */

package com.ibm.jikes.skij.misc;

import java.util.*;
import java.io.*;
import java.util.zip.*;

/**
 * This class implements a <code>ClassLoader</code> that allow to extend the
 * possible path to look up to load a class.
 *
 * @author      Olivier Dedieu
 * @version     0.1 Jan 28 1998 
 * @see         java.lang.ClassLoader
 * 
 */
public class ExtendableClassLoader extends ClassLoader {

  private Hashtable cache;
  public Vector paths;		// public for debug +++

  /**
   * Constructs a new class loader and initializes it.
   */
  public ExtendableClassLoader() {
    cache = new Hashtable();
    paths = new Vector();
  }

  /**
   * Add a path to the list of class paths
   *
   * @param path the class path
   */
  public void addClassPath(String path) {
    paths.addElement(path);
  }

  /**
   * Remove a class path
   * <P>
   * Remark: the previous loaded class with this path are not
   * unloaded.
   *
   * @param path the class path to remove
   */
  public void removeClassPath(String path) {
    for(Enumeration e = paths.elements(); e.hasMoreElements();) {
      String p = (String)e.nextElement();
      if (p.equals(path)) {
        paths.removeElement(p);
        break;
      }
    }
  }

  
  /**
   * Translate a class name into a file name.<P>
   * e.g: myPackage.MyClass -> myPackage/MyClass.class
   * <P>
   * Caution: it is assumed that the file separator is a one-character string
   *
   * @param className the class name to translate
   * @return the file name
   */
  private String ClassNameToFileName(String className) {
    return className.replace('.', System.getProperty("file.separator").charAt(0)) + ".class";
  }

  private String ClassNameToZipFileName(String className) {
    return className.replace('.', '/') + ".class";
  }


  /**
   * Looks up a class into all the class path and return its byte
   *  code.
   *
   * @param className the class to load
   * @return a byte array containing the byte code of the class
   */
  private byte[] getClassFromAddedClassPaths(String className) {

    String fsep = System.getProperty("file.separator");
    String fileName = ClassNameToFileName(className);
    String zipName = ClassNameToZipFileName(className);

    for(Enumeration enum = paths.elements(); enum.hasMoreElements();) {
      String path = (String)enum.nextElement();
      // System.out.println("Trying to find " + className + " in " + path);
      try {
	if (zipFileP(path)) {
	  ZipFile zip = new ZipFile(path);
	  ZipEntry entry = zip.getEntry(zipName);
 	  ZipInputStream zs = new ZipInputStream(new BufferedInputStream(new FileInputStream(path)));
	  // 	  ZipEntry entry = zs.createZipEntry(zipName);
 	  if (entry != null) {
	    return streamToBytes(zip.getInputStream(entry));
	  }
	}
	else {
	  File f = new File(path + fsep + fileName);
	  if (f.exists()) {
	    return streamToBytes(new FileInputStream(f));
	  }
	}
      }
      catch (Throwable e) {
	System.out.println("Error reading class " + className + " from " + path + ": " + e.toString());
      }
    }
    return null;
  }

  static byte[] streamToBytes(InputStream in) throws IOException {
    InputStream xin = new DataInputStream (new BufferedInputStream(in));
    byte[] result = new byte[in.available()];
    xin.read(result);
    return result;
  }

  static boolean zipFileP(String path) {
    return
      path.endsWith(".zip") ||
      path.endsWith(".jar");
  }

  /**
   * Requests the class loader to load a class with the specified
   * name. The loadClass method is called by the Java Virtual Machine
   * when a class loaded by a class loader first references another
   * class.
   *
   * @param name the name of the desired <code>Class</code>.add
   * @return the resulting <code>Class</code>, or <code>null</code> if
   * it was not found.
   * @exception ClassNotFoundException
   * if the class loader cannot find a definition for the class. 
   */
  public Class loadClass(String className) throws ClassNotFoundException {
    return (loadClass(className, true));
  }

  /**
   * Resolves the specified name to a Class. The method loadClass() is
   * called by the virtual machine.
   *
   * @param name the name of the desired <code>Class</code>.
   * @param resolve true if the <code>Class</code> needs to be resolved. 
   * @return the resulting <code>Class</code>, or <code>null</code> if
   * it was not found.
   * @exception ClassNotFoundException
   * if the class loader cannot find a definition for the class. 
   */
  public synchronized Class loadClass(String className, boolean resolveIt) throws ClassNotFoundException {
    Class result;
    byte classData[];

    // Check the cache of classes
    result = (Class)cache.get(className);
    if (result != null) {
      return result;
    }

    // Check with the primordial class loader 
    try {
      result = super.findSystemClass(className);
      return result;
    } catch (ClassNotFoundException e) {
    }

    // Try to load it from the added class paths
    classData = getClassFromAddedClassPaths(className);
    if (classData == null) {
      throw new ClassNotFoundException();
    }

    // Define it (parse the class file)
    result = defineClass(className, classData, 0, classData.length);
    if (result == null) {
      throw new ClassFormatError();
    }

    if (resolveIt) {
      resolveClass(result);
    }

    // Add the class to the cache
    cache.put(className, result);
    
    return result;
  }

  public Class xDefineClass(String name, byte[] bytes, int off, int len) {
    return defineClass(name, bytes, off, len);
  }
}


