package com.ibm.jikes.skij.util;
import java.lang.reflect.*;
import java.util.Vector;
import java.util.Hashtable;

/* This file is part of Skij.
 * Author: Michael Travers (mt@watson.ibm.com)
 * 
 * Licensed Materials - See the file license.txt.
 * (c) Copyright IBM Corp. 1997, 1998. All rights reserved.
 */

public class Invoke {

  static final boolean jdk12p = compareVersionStrings(System.getProperty("java.version"),"1.2");

  // This class implements the dynamic reflection capabilities of Skij.
  
  //// Method invocation

  // start with object and method name
  public static Object invoke(Object obj, String methodName, Object[] args) throws Throwable {
    return invoke(obj.getClass(), obj, methodName, args); 
  }
  
  // start with a class
  public static Object invoke(Class myClass, Object obj, String methodName, Object[] args) throws Throwable {
    Class[] signature = classArray(args);    
    Method method = lookupMethod(myClass, methodName, signature);
    if (method == null)
      throw new NoSuchMethodException("no applicable method");
    else {
      try {
	return method.invoke(obj, args); }
      // pass exceptions upward.
      catch (InvocationTargetException e) {
	throw(e.getTargetException()); }
    }
  }
  

  // analagous to invoke, but returns the method object instead of calling it.
  public static Object getMethod(Object obj, String methodName, Object[] args) throws Throwable {
    return getMethod(obj.getClass(), obj, methodName, args); 
  }
  
  // start with a class
  public static Method getMethod(Class myClass, Object obj, String methodName, Object[] args) throws NoSuchMethodException {
    Class[] signature = classArray(args);    
    return lookupMethod(myClass, methodName, signature);
  }


  /* Here is our arsenal of techniques:
     1) direct lookup (getMethod) (fast, but only works sometimes, and will cause an exception when it doesn't work)
     2) indirect lookup (slow)
     3) cached lookup (not that fast, because it has to create a key object)

     The strategy now is to try direct first, then cached, than indirect.  

     The alternative is (cache, direct, indirect), in other words, cache directs.  That may or may not be better.

     Also: right now indirect works through getMethods. Using getDirectMethods might be better.

     Note: if indirects are not cached, they should be less careless about calling trace functions.
   */


  // top-level -- try easy case first
  static Method lookupMethod(Class targetClass, String name, Class[] argClasses) throws NoSuchMethodException {  
    // first, see if there is an exact match
    try {
      return targetClass.getMethod(name, argClasses); }
    catch (NoSuchMethodException e) {	
      if (!jdk12p && argClasses.length == 0) {  // if no args and no exact match, we're out of luck
	return null; }
    }
    // go the more complicated route
    return lookupMethodCached(targetClass, name, argClasses); 
  }

  // +++ when JDK 1.2 comes out, use weak references
  static Hashtable hash = new Hashtable();

  static Method lookupMethodCached(Class target, String name, Class[] argClasses) throws NoSuchMethodException {
    MethodKey key = new MethodKey(target, name, argClasses);
    Object result = hash.get(key);
    if (result == null) {
      Method nresult = lookupMethodIndirect(target, name, argClasses);
      if (nresult == null)
	return null;
      else {
	hash.put(key, nresult);
	return nresult; }}
    else
      return (Method)result; }

  public static Method lookupMethodIndirect(Class targetClass, String methodName, Class[] argClasses) throws NoSuchMethodException {

    if (jdk12p) 
      return lookupMethodIndirect(new Vector(), targetClass, methodName, argClasses);
    else {
      Method[] methods = targetClass.getMethods();
      Vector goodMethods = new Vector();
      for (int i = 0; i != methods.length; i++) {
	if (methodName.equals(methods[i].getName()) &&
	    matchClasses(methods[i].getParameterTypes(), argClasses))
	  goodMethods.addElement(methods[i]);
      }
      switch (goodMethods.size()) {
      case 0: {
	return null; }
      case 1: {
	return (Method)goodMethods.firstElement(); }
      default: {
	return mostSpecificMethod(goodMethods);
      }}
    }
  }

  // jdk1.2 version 
  public static Method lookupMethodIndirect(Vector goodMethods, Class targetClass, String methodName, Class[] argClasses) throws NoSuchMethodException {

    if (targetClass == null) {
      Method m;
      switch (goodMethods.size()) {
      case 0: {
	return null; }
      case 1: {
	m = (Method)goodMethods.firstElement(); }
      default: {
	m = mostSpecificMethod(goodMethods);
      }}
      setAccessible(m, true);
      return m;
    }
    else {
      Method[] methods = targetClass.getDeclaredMethods();
      for (int i = 0; i != methods.length; i++) {
	Class[] paramTypes = methods[i].getParameterTypes();
	if (methodName.equals(methods[i].getName()) &&
	    matchClasses(paramTypes, argClasses))
	  goodMethods.addElement(methods[i]);
      }
      return lookupMethodIndirect(goodMethods, targetClass.getSuperclass(), methodName, argClasses);
    }
  }

  public static Object create(Class myClass, Object[] args) throws Throwable {
    Class[] signature = classArray(args);    
    Constructor[] constructors = jdk12p ? myClass.getDeclaredConstructors() : myClass.getConstructors();
    Vector goodConstructors = new Vector();
    for (int i = 0; i != constructors.length; i++) {
      if (matchClasses(constructors[i].getParameterTypes(), signature))
	goodConstructors.addElement(constructors[i]);
    }
    Constructor c;
    switch (goodConstructors.size()) {
    case 0: {
      throw new NoSuchMethodException("no applicable constructor"); }
    case 1: {
      c = (Constructor)goodConstructors.firstElement(); }
    default: {
      c = mostSpecific(goodConstructors);
    }
    }
    if (jdk12p) setAccessible(c, true);
    try {
      return c.newInstance(args);
    }
    catch (InvocationTargetException e) {
      throw(e.getTargetException()); 
    }
  }

  //// Peek and Poke

  public static void poke(Object object, String fieldName, Object value)
       throws IllegalAccessException, NoSuchFieldException {
	 Class myClass = object.getClass();
	 poke(object.getClass(), object, fieldName, value); 
  }

  public static void poke(Class myClass, Object object, String fieldName, Object value)
       throws IllegalAccessException, NoSuchFieldException {
	 Field field = fieldLookup(myClass, fieldName);
	 field.set(object, value); 
  }

  public static Object peek(Object object, String fieldName) 
       throws IllegalAccessException, NoSuchFieldException {
	 return peek(object.getClass(), object, fieldName);
  }

  public static Object peek(Class myClass, Object object, String fieldName) 
       throws IllegalAccessException, NoSuchFieldException {
	 Field field = fieldLookup(myClass, fieldName);
	 return field.get(object); 
  }

  static Field fieldLookup(Class myClass, String fieldName) throws NoSuchFieldException {
    Field field;
    try {
      field = myClass.getField(fieldName);	   
    }
    catch (NoSuchFieldException e) {
      if (jdk12p)
	field = fieldLookup0(myClass, fieldName);	// try harder
      else
	throw e;
    }
    return field;
  }

  // if it's a non-public field, use this uglier method
  static Field fieldLookup0(Class myClass, String fieldName) throws NoSuchFieldException {
    Field[] localFields = myClass.getDeclaredFields();
    for (int i = 0; i != localFields.length; i++) {
      if (fieldName.equals(localFields[i].getName())) {
	Field f = localFields[i];
	setAccessible(f, true);
	return f;
      }
    }
    // didn't find it, go up to superclass
    Class sup = myClass.getSuperclass();
    if (sup == null)
      throw new NoSuchFieldException(fieldName);
    else
      return fieldLookup(sup, fieldName);
  }


  // go thru reflection so I can compile this damned thing under 1.1
  static Method setAccessibleMethod = null;

  static void setAccessible(Object thing, boolean accessible) {
    try {
      if (setAccessibleMethod == null) {
	Class aclass = Class.forName("java.lang.reflect.AccessibleObject");
	setAccessibleMethod =
	  aclass.getMethod("setAccessible", new Class[]{ booleanClass });
      }
      setAccessibleMethod.invoke(thing, new Object[]{ Boolean.TRUE }); // +++ array can be cached
    }
    catch (Throwable e) {
      System.out.println("Error trying to set accessibility for " + thing);
    }
  }

  //// Utilities

  static boolean matchClasses(Class[] mclasses, Class[] pclasses) {
    if (mclasses.length == pclasses.length) {
      for (int i = 0; i != mclasses.length; i++) {
	if (!matchClass(mclasses[i], pclasses[i])) {
	  return false; }
      }
      return true;
    }
    return false;
  }

  static Class NullClass;
  static Class nullClass;
  static Class BooleanClass;
  static Class booleanClass;
  static Class CharacterClass;
  static Class characterClass;
  static Class ByteClass;
  static Class byteClass;
  static Class ShortClass;
  static Class shortClass;
  static Class IntegerClass;
  static Class integerClass;
  static Class LongClass;
  static Class longClass;
  static Class FloatClass;
  static Class floatClass;
  static Class DoubleClass;
  static Class doubleClass;

  // have to use static initializer because of exceptions
  static {
    try {
      nullClass = Void.TYPE;
      NullClass = Class.forName("java.lang.Void");
      booleanClass = Boolean.TYPE;
      BooleanClass = Class.forName("java.lang.Boolean");
      characterClass = Character.TYPE;
      CharacterClass = Class.forName("java.lang.Character");
      byteClass = Byte.TYPE;
      ByteClass = Class.forName("java.lang.Byte");
      shortClass = Short.TYPE;
      ShortClass = Class.forName("java.lang.Short");
      integerClass = Integer.TYPE;
      IntegerClass = Class.forName("java.lang.Integer");
      longClass = Long.TYPE;
      LongClass = Class.forName("java.lang.Long");
      floatClass = Float.TYPE;
      FloatClass = Class.forName("java.lang.Float");
      doubleClass = Double.TYPE;
      DoubleClass = Class.forName("java.lang.Double");
    }
    catch (ClassNotFoundException e) {
      System.out.println("this shouldn't happen");
    }
  }
      
  static boolean matchClass(Class mclass, Class pclass) {
    mclass = boxedClass(mclass);
    pclass = boxedClass(pclass);
    return 
      mclass == pclass ||
      mclass.isAssignableFrom(pclass) || 
      pclass == NullClass; // this may not be right for primitives?
  }


  static Class boxedClass(Class clazz) {
    if (clazz.isPrimitive()) {
      Class boxedClass;
      if (clazz == booleanClass)
	boxedClass = BooleanClass;
      else if (clazz == integerClass)
	boxedClass = IntegerClass;
      else if (clazz == characterClass)
	boxedClass = CharacterClass;
      else if (clazz == byteClass)
	boxedClass = ByteClass;
      else if (clazz == shortClass)
	boxedClass = ShortClass;
      else if (clazz == longClass)
	boxedClass = LongClass;
      else if (clazz == floatClass)
	boxedClass = FloatClass;
      else if (clazz == doubleClass)
	boxedClass = DoubleClass;
      else throw new Error("Unknown primitive type" + clazz);
      return boxedClass;
    }
    else 
      return clazz;
  }
  
  static Constructor mostSpecific(Vector constructors) throws Throwable {
    for (int i = 0; i != constructors.size(); i++) {
      for (int j = 0; j != constructors.size(); j++) {
	if ((i != j) &&
	    (moreSpecific((Constructor)constructors.elementAt(i), (Constructor)constructors.elementAt(j)))) {
	  constructors.removeElementAt(j);
	  if (i > j) i--;
	  j--;
	}
      }
    }
    if (constructors.size() == 1)
      return (Constructor)constructors.elementAt(0);
    else
      throw new NoSuchMethodException(">1 most specific constructor");
  }

  // true if c1 is more specific than c2
  static boolean moreSpecific(Constructor c1, Constructor c2) {
    if (matchClass(c1.getDeclaringClass(), c2.getDeclaringClass())) // needed for jdk12 only
      return false;
    Class[] p1 = c1.getParameterTypes();
    Class[] p2 = c2.getParameterTypes();
    int n = p1.length;
    for (int i = 0; i != n; i++) {
      if (matchClass(p1[i], p2[i])) {
	return false;
      }
    }
    return true;
  }

  // these are exactly the same as above except for types, which suggests (duh) that
  // Method and Class should have had a common ancestor or interface.
  static Method mostSpecificMethod(Vector methods) throws NoSuchMethodException {
    for (int i = 0; i != methods.size(); i++) {
      for (int j = 0; j != methods.size(); j++) {
	if ((i != j) &&
	    (moreSpecific((Method)methods.elementAt(i), (Method)methods.elementAt(j)))) {
	  methods.removeElementAt(j);
	  if (i > j) i--;
	  j--;
	}
      }
    }
    if (methods.size() == 1)
      return (Method)methods.elementAt(0);
    else
      throw new NoSuchMethodException(">1 most specific method");
  }

  // true if c1 is more specific than c2
  static boolean moreSpecific(Method c1, Method c2) {
    if (!matchClass(c2.getDeclaringClass(), c1.getDeclaringClass())) // needed for jdk12 only
      return false;
    Class[] p1 = c1.getParameterTypes();
    Class[] p2 = c2.getParameterTypes();
    int n = p1.length;
    for (int i = 0; i != n; i++) {
      if (!matchClass(p2[i], p1[i])) {
	return false;
      }
    }
    return true;
  }

  // given an array of objects, return an array of corresponding classes
  static Class[] classArray(Object[] args) {
    Class[] classes = new Class[args.length];
    for (int i = 0; i != args.length; i = i + 1)
      classes[i] = ((args[i] == null) ? NullClass : args[i].getClass());
    return classes; 
  }

  // Built-in array printing is lousy, so use this. should be in its own package. For debug only
  public static String toStringArray(Object[] array) {
    java.io.StringWriter out = new java.io.StringWriter();
    out.write('[');
    for (int i = 0; i != array.length; i++) {
      out.write(array[i].toString());
      out.write(' '); }
    out.write(']');
    return out.toString(); 
  }


  public static boolean compareVersionStrings(String test, String mustbe) {
    return compareVersionStrings(test, mustbe, 0, 0);
  }

  static boolean compareVersionStrings(String test, String must, int testPos, int mustPos) {
    int testV; int mustV;

    if (testPos >= test.length())
      testV = 0;
    else {
      int testNextPos = test.indexOf('.', testPos);
      if (testNextPos == -1) testNextPos = test.length();
      testV = parseInt(test.substring(testPos, testNextPos));
      testPos = testNextPos + 1;
    }
    if (mustPos >= must.length())
      mustV = 0;
    else {
      int mustNextPos = must.indexOf('.', mustPos);
      if (mustNextPos == -1) mustNextPos = must.length();
      mustV = parseInt(must.substring(mustPos, mustNextPos));
      mustPos = mustNextPos + 1;
    }

    if (testV < mustV)
      return false;
    else
      if (testV == mustV)
	if (testV == 0)
	  return true;
	else
	  return compareVersionStrings(test, must, testPos, mustPos);
      else
	return true;		// test > must
  }

  // like Integer.parseInt, but stops rather than errors when it hits a nonnumeric char
  static int parseInt(String str) {
    int end = str.length();
    for (int i = 0; i != end; i++) {
      if (Character.digit(str.charAt(i), 10) < 0) {
	end = i; break;
      }
    }
    return Integer.parseInt(str.substring(0, end));
  }



}

//// hash cache support

class MethodKey {
    Class target;
    String name;
    Class[] argClasses;

  public String toString() {
    return "<<" + target + "." + name + Invoke.toStringArray(argClasses) + ">>"; }

  MethodKey(Class t,String n, Class[] a) {
    target = t; name = n; argClasses = a; }

  public boolean equals(Object mx) {
    if (mx instanceof MethodKey) {
      MethodKey m = (MethodKey)mx;
      boolean v =
	(target == m.target &&
	 name.equals(m.name) &&
	 arrayEquals(argClasses, m.argClasses));
      return v; }
    else
      return false;
  }

  static boolean arrayEquals(Class[] c1, Class[] c2) {
    if (c1.length == c2.length) {
      for (int i = 0; i != c1.length; i++) {
	if (c1[i] != c2[i])
	  return false;}
      return true; }
    else
      return false;
  }


  public int hashCode() {
    int v = target.hashCode() ^ (37*name.hashCode());
    for (int i = 0; i != argClasses.length; i++) // added this, but seems to make things slower...
      v = (v * 43) ^ argClasses[i].hashCode();
    return v;
  }



}
